/*
             LUFA Library
     Copyright (C) Dean Camera, 2021.

  dean [at] fourwalledcubicle [dot] com
           www.lufa-lib.org
*/

/*
  Copyright 2021  Dean Camera (dean [at] fourwalledcubicle [dot] com)

  Permission to use, copy, modify, distribute, and sell this
  software and its documentation for any purpose is hereby granted
  without fee, provided that the above copyright notice appear in
  all copies and that both that the copyright notice and this
  permission notice and warranty disclaimer appear in supporting
  documentation, and that the name of the author not be used in
  advertising or publicity pertaining to distribution of the
  software without specific, written prior permission.

  The author disclaims all warranties with regard to this
  software, including all implied warranties of merchantability
  and fitness.  In no event shall the author be liable for any
  special, indirect or consequential damages or any damages
  whatsoever resulting from loss of use, data or profits, whether
  in an action of contract, negligence or other tortious action,
  arising out of or in connection with the use or performance of
  this software.
*/

/** \file
 *
 *  Main source file for the Minimal demo. This file contains the main tasks of the demo and
 *  is responsible for the initial application hardware configuration.
 */

#define  INCLUDE_FROM_MINIMAL_C
#define CYCLES 10000

#include "Minimal.h"
#include "util/delay.h"
#include "stdint.h"


// Configuration
struct Configuration {
    unsigned int level1_alert;
    unsigned int level2_alert;
    unsigned int level3_alert;
};

// LED info structure
struct Connector {
    int BIT;
    volatile unsigned char *DDR;
    volatile unsigned char *PORT;
};

// Notes
enum Note {
    DO = 261,
    RE = 293,
    MI = 329,
    FA = 349,
    SOL = 392,
    LA = 440,
    SI = 493,
    DO2 = 523,
    RE2 = 587,
    MI2 = 659,
    FA2 = 698,
    SOL2 = 784,
    LA2 = 880,
    SI2 = 987
};

// Declarations
#define NUMBER_OF_LEDS 8
#define DAC_SIZE 8
#define TEMPO 240

#define DEFAULT_CLKSEL0 0b00010101
#define DEFAULT_CLKSEL1 0b00001111
#define DEFAULT_CLKPR 0b10000000

#define ADFR 5
#define MAX_INT 256

//definir addresses de stockage dans l'eeprom
#define LEVEL1_ALERT_ADDRESS 0
#define LEVEL2_ALERT_ADDRESS 16
#define LEVEL3_ALERT_ADDRESS 32


void ad_init(unsigned char channel) {    // Sélectionne un canal
    ADCSRA |= (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0); // Division de fréquence 128 => 125KHz
    ADCSRA &= ~(1 << ADFR);                       // Mode conversion unique
    ADMUX |= (1 << REFS0) | (1 << ADLAR);             // Référence de mesure AVCC
    ADMUX = (ADMUX & 0xf0) | channel;             // Selection du canal
    ADCSRA |= (1 << ADEN);                      // Convertisseur activé
}

unsigned int ad_capture(void) {          // Acquisition de tension
    ADCSRA |= (1 << ADSC);                      // Début de conversion
    while (bit_is_set(ADCSRA, ADSC));        // Attente de la fin de conversion
    return ADCH;                            // Résultat sur 8 bits car ADLAR=1
}

/**
 * Ports initialisation
 * @param LED LED array
 * @param microphone Microphone
 * @param speaker Speaker array
 */
void init_elements(struct Connector *LED, struct Connector microphone, struct Connector *speaker) {
    // LED array initialisation
    // LED 0
    LED[0].BIT = 4;
    LED[0].DDR = &DDRD;
    LED[0].PORT = &PORTD;

    // LED 1
    LED[1].BIT = 6;
    LED[1].DDR = &DDRD;
    LED[1].PORT = &PORTD;

    // LED 2
    LED[2].BIT = 7;
    LED[2].DDR = &DDRD;
    LED[2].PORT = &PORTD;

    // LED 3
    LED[3].BIT = 4;
    LED[3].DDR = &DDRB;
    LED[3].PORT = &PORTB;

    // LED 4
    LED[4].BIT = 5;
    LED[4].DDR = &DDRB;
    LED[4].PORT = &PORTB;

    // LED 5
    LED[5].BIT = 6;
    LED[5].DDR = &DDRB;
    LED[5].PORT = &PORTB;

    // LED 6
    LED[6].BIT = 6;
    LED[6].DDR = &DDRC;
    LED[6].PORT = &PORTC;

    // LED 7
    LED[7].BIT = 7;
    LED[7].DDR = &DDRC;
    LED[7].PORT = &PORTC;

    // Microphone initialisation
    microphone.BIT = 0;
    microphone.DDR = &DDRF;
    microphone.PORT = &PORTF;

    *microphone.DDR |= ~(1 << microphone.BIT);

    // Speaker initialisation
    // Speaker 0
    speaker[0].BIT = 0;
    speaker[0].DDR = &DDRB;
    speaker[0].PORT = &PORTB;

    // Speaker 1
    speaker[1].BIT = 1;
    speaker[1].DDR = &DDRB;
    speaker[1].PORT = &PORTB;

    // Speaker 2
    speaker[2].BIT = 2;
    speaker[2].DDR = &DDRB;
    speaker[2].PORT = &PORTB;

    // Speaker 3
    speaker[3].BIT = 3;
    speaker[3].DDR = &DDRB;
    speaker[3].PORT = &PORTB;

    // Speaker 4
    speaker[4].BIT = 7;
    speaker[4].DDR = &DDRB;
    speaker[4].PORT = &PORTB;

    // Speaker 5
    speaker[5].BIT = 1;
    speaker[5].DDR = &DDRD;
    speaker[5].PORT = &PORTD;

    // Speaker 6
    speaker[6].BIT = 2;
    speaker[6].DDR = &DDRD;
    speaker[6].PORT = &PORTD;

    // Speaker 7
    speaker[7].BIT = 3;
    speaker[7].DDR = &DDRD;
    speaker[7].PORT = &PORTD;


    // Initialize DDR
    for (int i = 0; i < NUMBER_OF_LEDS; i++) {
        *LED[i].DDR |= (1 << LED[i].BIT);
    }

    for (int i = 0; i < DAC_SIZE; i++) {
        *speaker[i].DDR |= (1 << speaker[i].BIT);
    }
}

/**
 * Make a delay for a note, waiting for the mid period of the note
 * @param note Note to play
 */
void make_note_delay(enum Note note) {
    switch (note) {
        case DO:
            _delay_us(1000000 / 261 / 2);
            break;
        case RE:
            _delay_us(1000000 / 293 / 2);
            break;

        case MI:
            _delay_us(1000000 / 329 / 2);
            break;

        case FA:
            _delay_us(1000000 / 349 / 2);
            break;

        case SOL:
            _delay_us(1000000 / 392 / 2);
            break;

        case LA:
            _delay_us(1000000 / 440 / 2);
            break;

        case SI:
            _delay_us(1000000 / 493 / 2);
            break;

        case DO2:
            _delay_us(1000000 / 523 / 2);
            break;

        case RE2:
            _delay_us(1000000 / 587 / 2);
            break;

        case MI2:
            _delay_us(1000000 / 659 / 2);
            break;

        case FA2:
            _delay_us(1000000 / 698 / 2);
            break;

        case SOL2:
            _delay_us(1000000 / 784 / 2);
            break;

        case LA2:
            _delay_us(1000000 / 880 / 2);
            break;

        case SI2:

            _delay_us(1000000 / 987 / 2);
            break;
    }
}

/**
 * Play a note for a given time
 * @param speaker Speaker array
 * @param note Note to play
 * @param seconds Time to play the note
 */
void play_note(struct Connector *speaker, enum Note note, int seconds) {
    unsigned int period = (note * (60 / seconds)) / TEMPO;
    for (int i = 0; i < period / 2; i++) {
        for (int j = 0; j < DAC_SIZE; j++) {
            *speaker[j].PORT |= (1 << speaker[j].BIT);
        }

        make_note_delay(note);

        for (int j = 0; j < DAC_SIZE; j++) {
            *speaker[j].PORT &= ~(1 << speaker[j].BIT);
        }
        make_note_delay(note);
    }
}

/**
 * Play the welcome effect
 * @param speaker Speaker array
 * @param leds LED array
 */
void play_welcome(struct Connector *speaker, struct Connector *leds) {
    // Disable all LEDs
    for (int i = 0; i < NUMBER_OF_LEDS; i++) {
        *leds[i].PORT &= ~(1 << leds[i].BIT);
    }

    // Light up all LEDs one by one
    for (int i = 0; i < NUMBER_OF_LEDS; i++) {
        *leds[i].PORT |= (1 << leds[i].BIT);
        _delay_ms(100);
    }

    //array of notes
    enum Note notes[22] = {MI2, RE2, FA, FA, SOL, SOL, DO2, SI, RE, RE, MI,
                           MI, SI, LA, DO, DO, MI, MI, LA, LA, LA, LA};

    for (int i = 0; i < 21; i++) {
        play_note(speaker, notes[i], 1);
    }


    // Disable all LEDs one by one
    for (int i = 7; i >= 0; i--) {
        *leds[i].PORT &= ~(1 << leds[i].BIT);
        _delay_ms(50);
    }
}

/**
 * Read the sound and light up the LEDs
 * @param config Configuration
 * @param speaker Speaker data
 * @param leds LED array
 */
void read_sound(struct Configuration config, struct Connector *speaker, struct Connector *leds) {
    ad_init(0);
    unsigned int db = MAX_INT - ad_capture();
    for (int j = 0; j < NUMBER_OF_LEDS; j++) {
        *leds[j].PORT &= ~(1 << leds[j].BIT);
    }

    if (db > config.level3_alert) {
        for (int i = 0; i < NUMBER_OF_LEDS; i++) {
            *leds[i].PORT |= (1 << leds[i].BIT);
        }
        play_note(speaker, SI, 1);
        play_note(speaker, LA, 1);
        _delay_ms(50);
    }
    if (db > (config.level3_alert - config.level2_alert) * 2 / 3 + config.level2_alert) {
        *leds[1].PORT |= (1 << leds[1].BIT);
    }
    if (db > (config.level3_alert - config.level2_alert) * 1 / 3 + config.level2_alert) {
        *leds[2].PORT |= (1 << leds[2].BIT);
    }
    if (db > config.level2_alert) {
        *leds[3].PORT |= (1 << leds[3].BIT);
    }
    if (db > (config.level2_alert - config.level1_alert) * 2 / 3 + config.level1_alert) {
        *leds[4].PORT |= (1 << leds[4].BIT);
    }
    if (db > (config.level2_alert - config.level1_alert) * 1 / 3 + config.level1_alert) {
        *leds[5].PORT |= (1 << leds[5].BIT);
    }
    if (db > config.level1_alert) {
        *leds[6].PORT |= (1 << leds[6].BIT);
    }
    if (db > config.level1_alert / 2) {
        *leds[7].PORT |= (1 << leds[7].BIT);
    }
}

/**
 * Read data from USB
 */
void read_usb() {
    Endpoint_SelectEndpoint(ICHUT_OUT_EPADDR);

    if (Endpoint_IsOUTReceived()) {
        if (Endpoint_IsReadWriteAllowed()) {
            uint16_t stage1 = Endpoint_Read_16_LE();
            eeprom_update_word((uint16_t *) LEVEL1_ALERT_ADDRESS, stage1);

            uint16_t stage2 = Endpoint_Read_16_LE();
            eeprom_update_word((uint16_t *) LEVEL2_ALERT_ADDRESS, stage2);

            uint16_t stage3 = Endpoint_Read_16_LE();
            eeprom_update_word((uint16_t *) LEVEL3_ALERT_ADDRESS, stage3);
        }
    }
    Endpoint_ClearOUT();
}

/** Main program entry point. This routine configures the hardware required by the application, then
 *  enters a loop to run the application tasks in sequence.
 */
int main(void) {
    CLKSEL0 = DEFAULT_CLKSEL0;   // External clock selection
    CLKSEL1 = DEFAULT_CLKSEL1;   // 8Mhz minimum
    CLKPR = DEFAULT_CLKPR;       // Modification of the clock divider (CLKPCE = 1)
    CLKPR = 0;                   // 0 to disable the divider (divisor of 1)
    MCUCR |= (1 << JTD);           // Disable JTAG

    SetupHardware();
    GlobalInterruptEnable();

    struct Connector leds[NUMBER_OF_LEDS];
    struct Connector speaker[DAC_SIZE];
    struct Connector microphone;
    struct Configuration config;

    config.level1_alert = eeprom_read_word((uint16_t *) LEVEL1_ALERT_ADDRESS);
    config.level2_alert = eeprom_read_word((uint16_t *) LEVEL2_ALERT_ADDRESS);
    config.level3_alert = eeprom_read_word((uint16_t *) LEVEL3_ALERT_ADDRESS);

    // Initialization
    init_elements(leds, microphone, speaker);

    // Play the welcome effect
    play_welcome(speaker, leds);
    _delay_ms(50);

    int i = 0;

    while (1) {
        i++;
        if (i > CYCLES) {
            read_sound(config, speaker, leds);
            i = 0;
        } else {
            USB_USBTask();
            read_usb();
        }
    }
}


/**
 * Configures the board hardware and chip peripherals for the demo's functionality.
 */
void SetupHardware(void) {
#if (ARCH == ARCH_AVR8)
    /* Disable watchdog if enabled by bootloader/fuses */
    MCUSR &= ~(1 << WDRF);
    wdt_disable();

    /* Disable clock division */
    clock_prescale_set(clock_div_1);
#elif (ARCH == ARCH_XMEGA)
    /* Start the PLL to multiply the 2MHz RC oscillator to 32MHz and switch the CPU core to run from it */
    XMEGACLK_StartPLL(CLOCK_SRC_INT_RC2MHZ, 2000000, F_CPU);
    XMEGACLK_SetCPUClockSource(CLOCK_SRC_PLL);

    /* Start the 32MHz internal RC oscillator and start the DFLL to increase it to 48MHz using the USB SOF as a reference */
    XMEGACLK_StartInternalOscillator(CLOCK_SRC_INT_RC32MHZ);
    XMEGACLK_StartDFLL(CLOCK_SRC_INT_RC32MHZ, DFLL_REF_INT_USBSOF, F_USB);

    PMIC.CTRL = PMIC_LOLVLEN_bm | PMIC_MEDLVLEN_bm | PMIC_HILVLEN_bm;
#endif

    /* Hardware Initialization */
    USB_Init();
}

/** Event handler for the USB_Connect event. This indicates that the device is enumerating via the status LEDs. */
void EVENT_USB_Device_Connect(void) {
}

/** Event handler for the USB_Disconnect event. This indicates that the device is no longer connected to a host via
 *  the status LEDs.
 */
void EVENT_USB_Device_Disconnect(void) {
}

/** Event handler for the USB_ConfigurationChanged event. This is fired when the host set the current configuration
 *  of the USB device after enumeration - the device endpoints are configured.
 */
void EVENT_USB_Device_ConfigurationChanged(void) {
    Endpoint_ConfigureEndpoint(ICHUT_OUT_EPADDR, EP_TYPE_INTERRUPT, ICHUT_OUT_EPSIZE, 1);
}

/** Event handler for the USB_ControlRequest event. This is used to catch and process control requests sent to
 *  the device from the USB host before passing along unhandled control requests to the library for processing
 *  internally.
 */
void EVENT_USB_Device_ControlRequest(void) {
}
